unit CharViewMain;

// RJ CharView is released with full source code.
// The source code may be freely used and improved by anyone as long
// as the source code is released as well.
//
// v1.1 (2005-12-17)
// Added support for Unicode ranges between 0x10000 - 0x10FFFF
//
// v1.0 - First release by Rickard Johansson
// 2005-04-18.

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls, SMCombo, Buttons, Clipbrd, Grids,
  Contnrs, IniFiles;

type
  TForm1 = class(TForm)
    BtnAbout: TBitBtn;
    StatusBar1: TStatusBar;
    GroupBox1: TGroupBox;
    FontComboBox: TSMFontNamesCombo;
    StringGrid1: TStringGrid;
    GroupBox3: TGroupBox;
    BtnCopy: TBitBtn;
    Label1: TLabel;
    Label2: TLabel;
    GroupBox2: TGroupBox;
    btnClose: TBitBtn;
    btnAddChar: TBitBtn;
    CheckBoxHtml: TCheckBox;
    CheckBoxText: TCheckBox;
    btnClear: TBitBtn;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    txtHtml: TEdit;
    cboUnicode: TComboBox;
    cboCharset: TComboBox;
    Panel1: TPanel;
    Panel: TImage;
    Panel2: TPanel;
    imageText: TImage;
    procedure FormShow(Sender: TObject);
    procedure CheckBoxHtmlClick(Sender: TObject);
    procedure btnClearClick(Sender: TObject);
    procedure CheckBoxTextClick(Sender: TObject);
    procedure btnAddCharClick(Sender: TObject);
    procedure cboCharsetSelect(Sender: TObject);
    procedure StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure btnCloseClick(Sender: TObject);
    procedure FontComboBoxSelectionChange(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure cboUnicodeSelect(Sender: TObject);
    procedure StringGrid1DrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure FormCreate(Sender: TObject);
    procedure BtnCopyClick(Sender: TObject);
    procedure BtnAboutClick(Sender: TObject);
  private
    { Private declarations }
    FText: WideString;
    FClick: Boolean;
    FRanges: array of TSize;
    FBegin,
    FEnd: Integer;
    FChar: Integer;
    RangeIni: TMemIniFile;
    const CharsetArray: array[1..31] of Integer = (
      0,
      864, //DOS - Arabic
      775, //DOS - Baltic,
      855, //DOS - Cyrillic (primarily Russian)
      863, //DOS - French Canadian
      737, //DOS - Greek,
      856, //DOS - Hebrew
      861, //DOS - Icelandic
      862, //DOS - Israel
      869, //DOS - Modern Greek
      850, //DOS - Latin I (Western)
      852, //DOS - Latin II
      865, //DOS - Nordic
      860, //DOS - Portuguese
      866, //DOS - Russian
      857, //DOS - Turkish
      437, //DOS - USA
      1256, //Windows - Arabic
      1257, //Windows - Baltic
      1250, //Windows - Central European
      1251, //Windows - Cyrillic
      1253, //Windows - Greek
      1255, //Windows - Hebrew
      949,  //Windows - Korean
      932,  //Windows - Japanese
      936,  //Windows - Simplified Chinese
      874,  //Windows - Thai
      950,  //Windows - Traditional Chinese
      1254, //Windows - Turkish
      1258, //Windows - Vietnamese
      1252  //Windows - Western
      );
    procedure DisplayText;
  public
    function GetWideChar(n: Integer): WideString;
    procedure DisplayChar;
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses About;

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
var
  sz: string;
  i: Integer;
  Ini: TIniFile;
begin
  FChar := -1;
  FBegin := 0;
  FEnd := 127;

  StringGrid1.DoubleBuffered := True;
  sz := ExtractFileDir(Application.ExeName) + '\Data\Ranges.ini';
  RangeIni := TMemIniFile.Create(sz);

  if Assigned(RangeIni) then
  begin
    SetLength(FRanges,138);
    for i := 0 to 137 do
    begin
      FRanges[i].cx := RangeIni.ReadInteger('Ranges','0'+IntToStr(i),0);
      FRanges[i].cy := RangeIni.ReadInteger('Ranges','1'+IntToStr(i),0);
    end;
  end
  else
    ShowMessage('Could not load Unicode ranges');
end;

procedure TForm1.FormShow(Sender: TObject);
begin
  FontComboBox.FontName := TFontName('Arial');
  StringGrid1.Canvas.Font.Size := 10;
  StringGrid1.Canvas.Font.Color := clBlack;
  StringGrid1.Canvas.Brush.Color := clWhite;

  Panel.Canvas.Font.Color := clBlack;
  Panel.Canvas.Brush.Color := clWhite;

  imageText.Canvas.Font.Color := clBlack;
  imageText.Canvas.Brush.Color := clWhite;

  FontComboBoxSelectionChange(Sender);
  cboUnicode.ItemIndex := 0;
  cboUnicodeSelect(Sender);
end;

procedure TForm1.BtnCopyClick(Sender: TObject);
var
  Clip : TClipboard;
  mem : Cardinal;
  ptr : Pointer;
  SelText : WideString;
  SelLen : integer;
  S : String;
begin
  if CheckBoxHtml.Checked then
  begin
    Clipboard.Clear;
    Clipboard.AsText := txtHtml.Text;
    exit;
  end;
  Clip := Clipboard;
  Clip.Open;
  SelText := FText;
  SelLen := (Length(SelText) + 1) * 2;
  mem := GlobalAlloc(GMEM_MOVEABLE+GMEM_DDESHARE, SelLen);
  ptr := GlobalLock(mem);
  Move(PWideChar(SelText)^, ptr^, SelLen);
  GlobalUnlock(mem);
  Clip.SetAsHandle(CF_UNICODETEXT, mem);
  if Win32Platform <> VER_PLATFORM_WIN32_NT then
  begin
    S := SelText;
    SelLen := Length(S) + 1;
    mem := GlobalAlloc(GMEM_MOVEABLE + GMEM_DDESHARE, SelLen);
    ptr := GlobalLock(mem);
    Move(PChar(S)^, ptr^, SelLen);
    GlobalUnlock(mem);
    Clip.SetAsHandle(CF_TEXT, mem);
  end;

  Clip.Close;
end;

procedure TForm1.BtnAboutClick(Sender: TObject);
begin
  AsciiAbout.ShowModal;
end;

procedure TForm1.StringGrid1DrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  ws: WideString;
  x,y,n: Integer;
  size: TSize;
begin
  StringGrid1.Canvas.FillRect(Rect);

  n := FBegin + (16*ARow + ACol);
  if n > FEnd then exit;

  ws := GetWideChar(n);

  if gdSelected in State then
    StringGrid1.Canvas.DrawFocusRect(Rect);

  GetTextExtentPoint32W(StringGrid1.Canvas.Handle, PWideChar(ws), Length(ws), size);

  // Calculate position for the characters to place them in the
  // center of the cell.
  x := Rect.Left + ((Rect.Right - Rect.Left) - size.cx) div 2;
  y := Rect.Top + ((Rect.Bottom - Rect.Top) - size.cy) div 2;

  ExtTextOutW(StringGrid1.Canvas.Handle,x,y,ETO_CLIPPED,@Rect,PWideChar(ws),Length(ws),NIL);
end;

procedure TForm1.cboUnicodeSelect(Sender: TObject);
var
  n: Integer;
begin
  if cboCharset.Text <> 'UNICODE' then
  begin
    cboCharset.ItemIndex := 0;
  end;

  if not Assigned(FRanges) then
    exit;
  n := cboUnicode.ItemIndex;
  FBegin := FRanges[n].cx;
  FEnd := FRanges[n].cy;
  n := (FEnd - FBegin + 1) div 16;

  // Update the number of rows needed. Should be at least 16.
  if (n > 16) then
    StringGrid1.RowCount := n
  else
    StringGrid1.RowCount := 16;
  StringGrid1.Refresh;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  if Assigned(FRanges) then
  begin
    SetLength(FRanges,0);
    RangeIni.Free;
  end;
end;

procedure TForm1.FontComboBoxSelectionChange(Sender: TObject);
begin
  Panel.Canvas.Font.Name := FontComboBox.FontName;
  Panel.Canvas.Font.Size := 48;
  DisplayChar;
  imageText.Canvas.Font.Name := FontComboBox.FontName;
  imageText.Canvas.Font.Size := 10;
  DisplayText;
  StringGrid1.Canvas.Font.Name := FontComboBox.FontName;
  StringGrid1.Canvas.Font.Size := 10;
  StringGrid1.Refresh;
end;

procedure TForm1.btnCloseClick(Sender: TObject);
begin
  Close();
end;

procedure TForm1.StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer;
  var CanSelect: Boolean);
var
  n: Integer;
begin
  n := FBegin + ARow*16 + ACol;
  if n > FEnd then
    exit;

  FChar := n;
  DisplayChar;
end;

procedure TForm1.cboCharsetSelect(Sender: TObject);
begin
  if cboCharset.ItemIndex = 0 then
  begin
    cboUnicode.Enabled := True;
    cboUnicodeSelect(Sender);
    exit;
  end;
  cboUnicode.Enabled := False;
  FBegin := 0;
  FEnd := 255;
  FChar := -1;
  DisplayChar;
  StringGrid1.Refresh;
end;

function TForm1.GetWideChar(n: Integer): WideString;
var
  fc: Cardinal;
  c: AnsiChar;
  wc: WideChar;
begin
  Result := '';
  if cboCharset.Text <> 'UNICODE' then
  begin
    fc := CharsetArray[cboCharset.ItemIndex];
    if fc = 0 then
      fc := GetACP();
    c := AnsiChar(n);
    MultiByteToWideChar(fc, MB_PRECOMPOSED or MB_USEGLYPHCHARS, @c, 1, @wc, 1);
    Result := Result + wc;
  end
  else
  begin
    if (n < 65536) then
    begin
      // n < 0x10000
      Result := WideChar(n);
    end
    else  // n  0x10000 - 0x10FFFF
    begin
      // To display characters beyond 0xFFFF we have to use surrogate
      // pairs. A high surrogate is must be in the 0xD800 - 0xDBFF range
      // and the low surrogate in the 0xDC00 - 0xDFFF range.
      // So, the value 0x10000 (65536) has to be replaced with
      // 0xD800 DC00, 0x10001 with 0xD800 DC01 and so on...
      n := n - 65536;
      Result := Result + WideChar((n shr 10) + 55296) +
                         WideChar((n and 1023) + 56320);
    end;
  end;

end;

procedure TForm1.btnAddCharClick(Sender: TObject);
var
  w: WideString;
begin
  if FChar >= 0 then
  begin
    // Add to the text string
    FText := FText + GetWideChar(FChar);
    DisplayText;

    // Add to the html special characters string
    w := GetWideChar(FChar);

    w := '&#' + IntToStr(FChar) + ';';
    txtHtml.Text := txtHtml.Text + w;
  end;
end;

procedure TForm1.CheckBoxTextClick(Sender: TObject);
begin
  if FClick then
  begin
    FClick := False;
    exit;
  end;
  FClick := True;
  CheckBoxHtml.Checked := False;
end;

procedure TForm1.CheckBoxHtmlClick(Sender: TObject);
begin
  if FClick then
  begin
    FClick := False;
    exit;
  end;
  FClick := True;
  CheckBoxText.Checked := False;
end;

procedure TForm1.btnClearClick(Sender: TObject);
begin
  FText := '';
  txtHtml.Text := '';
  DisplayText;
end;

procedure TForm1.DisplayChar;
var
  ws: WideString;
  x,y: Integer;
  Rect: TRect;
  size: TSize;
begin
  Rect := Panel.ClientRect;
  Panel.Canvas.FillRect(Rect);

  if FChar < 0 then
    exit;

  ws := GetWideChar(FChar);
  GetTextExtentPoint32W(Panel.Canvas.Handle, PWideChar(ws), Length(ws), size);

  // Calculate the center position
  x := Rect.Left + ((Rect.Right - Rect.Left) - size.cx) div 2;
  y := Rect.Top + ((Rect.Bottom - Rect.Top) - size.cy) div 2;

  ExtTextOutW(Panel.Canvas.Handle,x,y,ETO_CLIPPED,@Rect,PWideChar(ws),Length(ws),NIL);
  Panel.Refresh;

  StatusBar1.SimpleText := 'U+' + IntToHex(FChar,8) + ', Html: &#' + IntToStr(FChar) + ';';
end;

procedure TForm1.DisplayText;
var
  ws: WideString;
  Rect: TRect;
  y: Integer;
  size: TSize;
begin
  ws := 'Tg';
  Rect := imageText.ClientRect;
  GetTextExtentPoint32W(imageText.Canvas.Handle, PWideChar(ws), 2, size);

  // Center the text
  y := Rect.Top + ((Rect.Bottom - Rect.Top) - size.cy) div 2;

  imageText.Canvas.FillRect(Rect);
  ExtTextOutW(imageText.Canvas.Handle,0,y,ETO_CLIPPED,@Rect,PWideChar(FText),Length(FText),NIL);
  imageText.Refresh;
end;

end.
